#ifdef VERTEX_SHADER

layout(location = 0) in vec3 a_position;
layout(location = 1) in vec3 a_normal;
layout(location = 2) in vec3 a_tangent;
layout(location = 3) in vec2 a_uv;

layout (std140, binding = 1) uniform Camera {
    mat4 view;
    mat4 view_inv;
    mat4 projection;
    mat4 projection_inv;
    mat4 viewProjection;
    vec3 camera_position;
    vec4 fov_near_far;
    uvec4 resolution;
    float time;
    vec2 padding;
    float relief_factor;
};

layout(std430, binding=5) buffer Model_in {
	mat4 model_matrix[];
};

out vec3 f_position;
out vec3 f_normal;
out vec3 f_tangent;
out vec2 f_uv;

void main()
{
    f_position = (model_matrix[gl_InstanceID] * vec4(a_position, 1.0)).xyz;
    f_normal = (model_matrix[gl_InstanceID] * vec4(a_normal, 0.0)).xyz;
    f_tangent = (model_matrix[gl_InstanceID] * vec4(a_tangent, 0.0)).xyz;
    f_uv = a_uv;

    gl_Position = viewProjection * vec4(f_position, 1.0);
}

#endif

#ifdef FRAGMENT_SHADER
#define NUM_OF_POINTLIGHTS 3
#define NUM_OF_SHADOW_SAMPLES 16

layout (std140, binding = 1) uniform Camera {
    mat4 view;
    mat4 view_inv;
    mat4 projection;
    mat4 projection_inv;
    mat4 viewProjection;
    vec3 camera_position;
    vec4 fov_near_far;
    uvec4 resolution;
    float time;
    vec2 padding;
    float relief_factor;
};

layout (std140, binding = 2) uniform DirectionalLight {
    vec3 direction;
    vec3 color;
} directionalLight;

struct PointLight{
    vec3 position;
    vec3 color;
    vec3 attenuation;
    vec2 near_far_plane;
    vec2 padding;
    vec2 tau_phi;
};

layout (std140, binding = 3) uniform PointLights {
    PointLight pointLights[NUM_OF_POINTLIGHTS];
};

in vec3 f_position;
in vec3 f_normal;
in vec3 f_tangent;
in vec2 f_uv;

uniform sampler2D diffuse_map;
uniform sampler2D normal_map;
uniform sampler2D depth_map;
uniform sampler2D specular_map;
uniform samplerCube shadow_map[NUM_OF_POINTLIGHTS];
uniform vec4 materialCoefficients; // x = ambient, y = diffuse, z = specular, w = specular alpha


vec3 shadow_samples[16] = vec3[](
vec3(0.382390,-0.763971,-0.519737),vec3(-0.476527,-0.210767,0.853522),vec3(-0.650877,0.298005,-0.698249),vec3(0.537643,-0.747269,0.390550),
vec3(0.780838,-0.188870,0.595500),vec3(0.611989,0.729557,-0.305313),vec3(0.893255,-0.370405,0.254748),vec3(-0.642904,0.765648,0.021419),
vec3(0.076607,0.663683,0.744081),vec3(-0.635874,-0.738714,0.223529),vec3(0.735506,-0.471653,-0.486389),vec3(0.632164,0.686413,0.359453),
vec3(0.920170,0.320219,-0.225270),vec3(-0.425501,-0.860727,0.279461),vec3(-0.435288,0.865974,-0.246196),vec3(-0.654006,-0.526157,0.543539)
);

layout (location = 0) out vec4 FragColor;
layout (location = 1) out vec4 BrightColor;

vec3 blinn_phong(vec3 n, vec3 l, vec3 v, vec3 albedo, vec3 spec_term, vec4 materialCoeff, vec3 lightColor, bool attenuate, vec3 attenuation)
{
    float d = length(l);
    l = normalize(l);

    float att = 1.0;	
	if(attenuate) {
        //att = 1.0 / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
        att = 1.0 / pow((attenuation.x + d * attenuation.y + d * d * attenuation.z), 2);
    }
    vec3 h = normalize(l + v);
    // diffuse + specular
    return (materialCoeff.y * albedo * lightColor * max(0, dot(n, l))) * att +
           (materialCoeff.z * lightColor * pow(max(0, dot(n, h)), materialCoeff.w)) * att * spec_term;
}

// shadow map function
float is_lit(vec3 w_position, int i) {
    vec3 light_to_pos = w_position - pointLights[i].position;
    float light_to_pos_length = length(light_to_pos);
    float bias = 0.05;
    float radius = 0.07;
    float lit = 0.0;
    for(uint n = 0; n < NUM_OF_SHADOW_SAMPLES; n++){
        float depth_dist = texture(shadow_map[i], light_to_pos + shadow_samples[n] * radius).r * pointLights[i].near_far_plane.y;
        lit += light_to_pos_length - bias > depth_dist ? 0.0 : 1.0;
    }
    return lit / NUM_OF_SHADOW_SAMPLES;
}

mat3 tangent_space_transform() {
    vec3 n = normalize(f_normal);
    vec3 t = f_tangent;
    t = normalize(t - dot(n, t) * n); // Gram-Schmidt re-orthogonalization
    return mat3(t, cross(n, t), n);
}

vec2 parallax_mapping(in vec3 view_dir_in_world_space, in mat3 TBN)
{
    float parallax_depth = 0.22;

    vec3 direction = transpose(TBN) * -view_dir_in_world_space; // Ray direction tangent space
    // Clamp Z
    //float z = -direction.z;
    //direction.z = mix(0.15, z, smoothstep(0.0, 0.6, z));
    direction.z = max(0.001, -direction.z);

    float texture_depth = texture(depth_map, f_uv).r * parallax_depth;
    return f_uv + direction.xy * texture_depth * parallax_depth / direction.z;
}

vec2 relief_mapping(in vec2 start_uv, 
                    in float parallax_depth, 
                    in sampler2D depth_map,
                    in vec3 view_dir_in_world_space, 
                    in mat3 TBN)
{
    float linear_step_count = 32.0;
    int binary_search_step_count = 6;
    vec3 origin = vec3(start_uv, 0); // Ray origin in tangent space
    vec3 direction = transpose(TBN) * -view_dir_in_world_space; // Ray direction tangent space
    // Smooth z so it does not go to 0 at very steep angles
    float z = -direction.z;
    direction.z = mix(0.05, z, smoothstep(0.0, 0.2, z));

    float t = 0.0;
    float dt = parallax_depth / linear_step_count / direction.z;

    vec3 p = origin;
    float texture_depth = texture(depth_map, p.xy).r * parallax_depth;
    while (p.z < texture_depth)
    {
        t += dt;
        p = origin + t * direction;
        texture_depth = texture(depth_map, p.xy).r * parallax_depth;
    }

    // Refine last intersection using binary search
    float inside_t = t;
    float outside_t = t - dt;
    for (int i = 0; i < binary_search_step_count; i++) 
    {
        float middle_t = (inside_t + outside_t)/2.0;
        p = origin + middle_t * direction;
        texture_depth = texture(depth_map, p.xy).r * parallax_depth;
        if (p.z > texture_depth) {
            inside_t = middle_t;
        }
        else {
            outside_t = middle_t;
        }
    }

    return p.xy;
}

vec2 relief_mapping_with_cutoff(
    in vec2 start_uv, 
    in float parallax_depth, 
    in sampler2D depth_map,
    in vec3 view_dir_in_world_space, 
    in mat3 TBN,
    in float dist,
    in float cutoff_start,
    in float cutoff_end)
{
    if (dist > cutoff_end) {
        return start_uv;
    }
    vec2 relief_uv = relief_mapping(start_uv, parallax_depth, depth_map, view_dir_in_world_space, TBN);
    return mix(relief_uv, start_uv, smoothstep(cutoff_start, cutoff_end, dist));
}

void main()
{
    // Get necessary vectors for light calculation
    vec3 v = normalize(camera_position - f_position);
    mat3 TBN = tangent_space_transform();
    vec2 uv = f_uv;
    uv = relief_mapping(uv, relief_factor, depth_map, v, TBN); // Relief mapping 
    //uv = relief_mapping_with_cutoff(uv, 0.10, depth_map, v, TBN, distance(camera_position, f_position), 25, 50);
    vec3 albedo = texture(diffuse_map, uv).rgb;
    vec3 spec_term = texture(specular_map, uv).rgb;
    vec3 n = f_normal;
    n = normalize(TBN * (2.0 * texture(normal_map, uv).xyz - vec3(1.0))); // Normal mapping

    // Light Calculation
    // ambient light
    FragColor = vec4(albedo * materialCoefficients.x, 1.0);

    // dir light
    //FragColor.rgb += blinn_phong(n, -directionalLight.direction, v, albedo, spec_term, materialCoefficients, directionalLight.color, false, vec3(0));

    // Point Lights
    for(int i = 0; i < NUM_OF_POINTLIGHTS; i++) {
        //int i = 0;
        FragColor.rgb += is_lit(f_position, i) * blinn_phong(n, pointLights[i].position - f_position, v, albedo, spec_term, materialCoefficients, pointLights[i].color, true, pointLights[i].attenuation);
    }

    // Generate bright color for blur if pixel intensity > 1
    float greyscale = dot(FragColor.rgb, vec3(0.2126, 0.7152, 0.0722));
    if(greyscale > 1.0) {
        BrightColor = vec4(FragColor.rgb, 1.0);
    }
    else {
        BrightColor = vec4(0.0, 0.0, 0.0, 1.0);
    }
}

#endif